PANDAS_TOCSV_KWARGS = [
    "path_or_buf",
    "sep",
    "na_rep",
    "float_format",
    "columns",
    "header",
    "index",
    "index_label",
    "mode",
    "encoding",
    "compression",
    "quoting",
    "quotechar",
    "line_terminator",
    "chunksize",
    "tupleize_cols",
    "date_format",
    "doublequote",
    "escapechar",
    "decimal",
]

PANDAS_READ_TABLE_KWARGS = [
    "sep",
    "delimiter",
    "header",
    "names",
    "index_col",
    "usecols",
    "squeeze",
    "prefix",
    "mangle_dupe_cols",
    "dtype",
    "engine",
    "converters",
    "true_values",
    "false_values",
    "skipinitialspace",
    "skiprows",
    "nrows",
    "na_values",
    "keep_default_na",
    "na_filter",
    "verbose",
    "skip_blank_lines",
    "parse_dates",
    "infer_datetime_format",
    "keep_date_col",
    "date_parser",
    "dayfirst",
    "iterator",
    "chunksize",
    "compression",
    "thousands",
    "decimal",
    "lineterminator",
    "quotechar",
    "quoting",
    "escapechar",
    "comment",
    "encoding",
    "dialect",
    "tupleize_cols",
    "error_bad_lines",
    "warn_bad_lines",
    "skipfooter",
    "doublequote",
    "delim_whitespace",
    "low_memory",
    "memory_map",
    "float_precision",
]
